# -*- coding: utf-8 -*-
"""The attribute container store interface."""

import abc
import collections

from acstore.containers import manager as containers_manager


class AttributeSerializer(object):
  """Attribute serializer."""

  @abc.abstractmethod
  def DeserializeValue(self, value):
    """Deserializes a value.

    Args:
      value (object): serialized value.

    Returns:
      object: runtime value.
    """

  @abc.abstractmethod
  def SerializeValue(self, value):
    """Serializes a value.

    Args:
      value (object): runtime value.

    Returns:
      object: serialized value.
    """


class AttributeContainerStore(object):
  """Interface of an attribute container store.

  Attributes:
    format_version (int): storage format version.
  """

  def __init__(self):
    """Initializes an attribute container store."""
    super(AttributeContainerStore, self).__init__()
    self._attribute_container_sequence_numbers = collections.Counter()
    self._containers_manager = containers_manager.AttributeContainersManager
    self._storage_profiler = None

    self.format_version = None

  def _GetAttributeContainerNextSequenceNumber(self, container_type):
    """Retrieves the next sequence number of an attribute container.

    Args:
      container_type (str): attribute container type.

    Returns:
      int: next sequence number.
    """
    self._attribute_container_sequence_numbers[container_type] += 1
    return self._attribute_container_sequence_numbers[container_type]

  def _GetAttributeContainerSchema(self, container_type):
    """Retrieves the schema of an attribute container.

    Args:
      container_type (str): attribute container type.

    Returns:
      dict[str, str]: attribute container schema or an empty dictionary if
          no schema available.
    """
    try:
      schema = self._containers_manager.GetSchema(container_type)
    except ValueError:
      schema = {}

    return schema

  @abc.abstractmethod
  def _RaiseIfNotReadable(self):
    """Raises if the store is not readable.

    Raises:
      OSError: if the store cannot be read from.
      IOError: if the store cannot be read from.
    """

  @abc.abstractmethod
  def _RaiseIfNotWritable(self):
    """Raises if the store is not writable.

    Raises:
      OSError: if the store cannot be written to.
      IOError: if the store cannot be written to.
    """

  def _SetAttributeContainerNextSequenceNumber(
      self, container_type, next_sequence_number):
    """Sets the next sequence number of an attribute container.

    Args:
      container_type (str): attribute container type.
      next_sequence_number (int): next sequence number.
    """
    self._attribute_container_sequence_numbers[
        container_type] = next_sequence_number

  @abc.abstractmethod
  def _WriteExistingAttributeContainer(self, container):
    """Writes an existing attribute container to the store.

    Args:
      container (AttributeContainer): attribute container.
    """

  @abc.abstractmethod
  def _WriteNewAttributeContainer(self, container):
    """Writes a new attribute container to the store.

    Args:
      container (AttributeContainer): attribute container.
    """

  def AddAttributeContainer(self, container):
    """Adds a new attribute container.

    Args:
      container (AttributeContainer): attribute container.

    Raises:
      OSError: if the store cannot be written to.
      IOError: if the store cannot be written to.
    """
    self._RaiseIfNotWritable()
    self._WriteNewAttributeContainer(container)

  @abc.abstractmethod
  def Close(self):
    """Closes the store."""

  @abc.abstractmethod
  def GetAttributeContainerByIdentifier(self, container_type, identifier):
    """Retrieves a specific type of container with a specific identifier.

    Args:
      container_type (str): container type.
      identifier (AttributeContainerIdentifier): attribute container identifier.

    Returns:
      AttributeContainer: attribute container or None if not available.
    """

  @abc.abstractmethod
  def GetAttributeContainerByIndex(self, container_type, index):
    """Retrieves a specific attribute container.

    Args:
      container_type (str): attribute container type.
      index (int): attribute container index.

    Returns:
      AttributeContainer: attribute container or None if not available.
    """

  @abc.abstractmethod
  def GetAttributeContainers(self, container_type, filter_expression=None):
    """Retrieves a specific type of attribute containers.

    Args:
      container_type (str): attribute container type.
      filter_expression (Optional[str]): expression to filter the resulting
          attribute containers by.

    Returns:
      generator(AttributeContainer): attribute container generator.
    """

  @abc.abstractmethod
  def GetNumberOfAttributeContainers(self, container_type):
    """Retrieves the number of a specific type of attribute containers.

    Args:
      container_type (str): attribute container type.

    Returns:
      int: the number of containers of a specified type.
    """

  @abc.abstractmethod
  def HasAttributeContainers(self, container_type):
    """Determines if a store contains a specific type of attribute container.

    Args:
      container_type (str): attribute container type.

    Returns:
      bool: True if the store contains the specified type of attribute
          containers.
    """

  @abc.abstractmethod
  def Open(self, **kwargs):
    """Opens the store."""

  def SetStorageProfiler(self, storage_profiler):
    """Sets the storage profiler.

    Args:
      storage_profiler (StorageProfiler): storage profiler.
    """
    self._storage_profiler = storage_profiler

  def UpdateAttributeContainer(self, container):
    """Updates an existing attribute container.

    Args:
      container (AttributeContainer): attribute container.

    Raises:
      OSError: if the store cannot be written to.
      IOError: if the store cannot be written to.
    """
    self._RaiseIfNotWritable()
    self._WriteExistingAttributeContainer(container)


class AttributeContainerStoreWithReadCache(AttributeContainerStore):
  """Interface of an attribute container store with read cache.

  Attributes:
    format_version (int): storage format version.
  """

  # pylint: disable=abstract-method

  # The maximum number of cached attribute containers
  _MAXIMUM_CACHED_CONTAINERS = 32 * 1024

  def __init__(self):
    """Initializes an attribute container store with read cache."""
    super(AttributeContainerStoreWithReadCache, self).__init__()
    self._attribute_container_cache = collections.OrderedDict()

  def _CacheAttributeContainerByIndex(self, attribute_container, index):
    """Caches a specific attribute container.

    Args:
      attribute_container (AttributeContainer): attribute container.
      index (int): attribute container index.
    """
    if len(self._attribute_container_cache) >= self._MAXIMUM_CACHED_CONTAINERS:
      self._attribute_container_cache.popitem(last=True)

    lookup_key = f'{attribute_container.CONTAINER_TYPE:s}.{index:d}'
    self._attribute_container_cache[lookup_key] = attribute_container
    self._attribute_container_cache.move_to_end(lookup_key, last=False)

  def _GetCachedAttributeContainer(self, container_type, index):
    """Retrieves a specific cached attribute container.

    Args:
      container_type (str): attribute container type.
      index (int): attribute container index.

    Returns:
      AttributeContainer: attribute container or None if not available.
    """
    lookup_key = f'{container_type:s}.{index:d}'
    attribute_container = self._attribute_container_cache.get(lookup_key, None)
    if attribute_container:
      self._attribute_container_cache.move_to_end(lookup_key, last=False)
    return attribute_container



# -*- coding: utf-8 -*-
"""The attribute container interface."""


class AttributeContainerIdentifier(object):
  """The attribute container identifier.

  The identifier is used to uniquely identify attribute containers.
  The value should be unique relative to an attribute container store.

  Attributes:
    name (str): name of the table (attribute container).
    sequence_number (int): sequence number of the attribute container.
  """

  def __init__(self, name=None, sequence_number=None):
    """Initializes an attribute container identifier.

    Args:
      name (Optional[str]): name of the table (attribute container).
      sequence_number (Optional[int]): sequence number of the attribute
          container.
    """
    super(AttributeContainerIdentifier, self).__init__()
    self.name = name
    self.sequence_number = sequence_number

  def CopyFromString(self, identifier_string):
    """Copies the identifier from a string representation.

    Args:
      identifier_string (str): string representation.
    """
    self.name, sequence_number = identifier_string.split('.')
    self.sequence_number = int(sequence_number, 10)

  def CopyToString(self):
    """Copies the identifier to a string representation.

    Returns:
      str: unique identifier or None.
    """
    if self.name is not None and self.sequence_number is not None:
      return f'{self.name:s}.{self.sequence_number:d}'

    return None


class AttributeContainer(object):
  """The attribute container interface.

  This is the base class for those object that exists primarily as a container
  of attributes with basic accessors and mutators.

  The CONTAINER_TYPE class attribute contains a string that identifies the
  container type, for example the container type "event" identifiers an event
  object.

  Attributes are public class members of a serializable type. Protected and
  private class members are not to be serialized, with the exception of those
  defined in _SERIALIZABLE_PROTECTED_ATTRIBUTES.
  """

  CONTAINER_TYPE = None

  # Names of protected attributes, those with a leading underscore, that
  # should be serialized.
  _SERIALIZABLE_PROTECTED_ATTRIBUTES = []

  def __init__(self):
    """Initializes an attribute container."""
    super(AttributeContainer, self).__init__()
    self._identifier = AttributeContainerIdentifier(
        name=self.CONTAINER_TYPE, sequence_number=id(self))

  def CopyFromDict(self, attributes):
    """Copies the attribute container from a dictionary.

    Args:
      attributes (dict[str, object]): attribute values per name.
    """
    for attribute_name, attribute_value in attributes.items():
      # Not using startswith to improve performance.
      if (attribute_name[0] != '_' or
          attribute_name in self._SERIALIZABLE_PROTECTED_ATTRIBUTES):
        self.__dict__[attribute_name] = attribute_value

  def CopyToDict(self):
    """Copies the attribute container to a dictionary.

    Returns:
      dict[str, object]: attribute values per name.
    """
    return dict(self.GetAttributes())

  def GetAttributeNames(self):
    """Retrieves the names of all attributes.

    Returns:
      list[str]: attribute names.
    """
    attribute_names = list(self._SERIALIZABLE_PROTECTED_ATTRIBUTES)
    for attribute_name in self.__dict__:
      # Not using startswith to improve performance.
      if attribute_name[0] != '_':
        attribute_names.append(attribute_name)

    return attribute_names

  def GetAttributes(self):
    """Retrieves the attribute names and values.

    Attributes that are set to None are ignored.

    Yields:
      tuple[str, object]: attribute name and value.
    """
    for attribute_name, attribute_value in self.__dict__.items():
      # Not using startswith to improve performance.
      if attribute_value is not None and (
          attribute_name[0] != '_' or
          attribute_name in self._SERIALIZABLE_PROTECTED_ATTRIBUTES):
        yield attribute_name, attribute_value

  def GetAttributeValuesHash(self):
    """Retrieves a comparable string of the attribute values.

    Returns:
      int: hash of comparable string of the attribute values.
    """
    return hash(self.GetAttributeValuesString())

  def GetAttributeValuesString(self):
    """Retrieves a comparable string of the attribute values.

    Returns:
      str: comparable string of the attribute values.
    """
    attributes = []
    for attribute_name, attribute_value in sorted(self.__dict__.items()):
      # Not using startswith to improve performance.
      if attribute_value is not None and (
          attribute_name[0] != '_' or
          attribute_name in self._SERIALIZABLE_PROTECTED_ATTRIBUTES):
        if isinstance(attribute_value, dict):
          attribute_value = sorted(attribute_value.items())

        elif isinstance(attribute_value, bytes):
          attribute_value = repr(attribute_value)

        attributes.append(f'{attribute_name:s}: {attribute_value!s}')

    return ', '.join(attributes)

  def GetIdentifier(self):
    """Retrieves the identifier.

    The identifier is a storage specific value that should not be serialized.

    Returns:
      AttributeContainerIdentifier: a unique identifier for the container.
    """
    return self._identifier

  def MatchesExpression(self, expression):
    """Determines if an attribute container matches the expression.

    Args:
      expression (code|str): expression.

    Returns:
      bool: True if the attribute container matches the expression, False
          otherwise.
    """
    result = not expression
    if expression:
      namespace = {}
      for attribute_name, attribute_value in self.__dict__.items():
        # Not using startswith to improve performance.
        if attribute_value is not None and (
            attribute_name[0] != '_' or
            attribute_name in self._SERIALIZABLE_PROTECTED_ATTRIBUTES):
          if isinstance(attribute_value, AttributeContainerIdentifier):
            attribute_value = attribute_value.CopyToString()

          namespace[attribute_name] = attribute_value

      # Make sure __builtins__ contains an empty dictionary.
      namespace['__builtins__'] = {}

      try:
        result = eval(expression, namespace)  # pylint: disable=eval-used
      except Exception:  # pylint: disable=broad-except
        pass

    return result

  def SetIdentifier(self, identifier):
    """Sets the identifier.

    The identifier is a storage specific value that should not be serialized.

    Args:
      identifier (AttributeContainerIdentifier): identifier.
    """
    self._identifier = identifier



# -*- coding: utf-8 -*-
"""Schema helper."""


class SchemaHelper(object):
  """Schema helper."""

  # Data types and corresponding attribute serializers per method.
  _data_types = {
      'AttributeContainerIdentifier': None,
      'bool': None,
      'int': None,
      'str': None,
      'timestamp': None}

  @classmethod
  def DeregisterDataType(cls, data_type):
    """Deregisters a data type.

    Args:
      data_type (str): data type.

    Raises:
      KeyError: if the data type is not set.
    """
    if data_type not in cls._data_types:
      raise KeyError(f'Data type: {data_type:s} not set.')

    del cls._data_types[data_type]

  @classmethod
  def GetAttributeSerializer(cls, data_type, serialization_method):
    """Retrieves a specific attribute serializer.

    Args:
      data_type (str): data type.
      serialization_method (str): serialization method.

    Returns:
      AttributeSerializer: attribute serializer or None if not available.
    """
    serializers = cls._data_types.get(data_type, None) or {}
    return serializers.get(serialization_method, None)

  @classmethod
  def HasDataType(cls, data_type):
    """Determines is a specific data type is supported by the schema.

    Args:
      data_type (str): data type.

    Returns:
      bool: True if the data type is supported, or False otherwise.
    """
    return data_type in cls._data_types

  @classmethod
  def RegisterDataType(cls, data_type, serializers):
    """Registers a data type.

    Args:
      data_type (str): data type.
      serializers (dict[str, AttributeSerializer]): attribute serializers per
          method.

    Raises:
      KeyError: if the data type is already set.
    """
    if data_type in cls._data_types:
      raise KeyError(f'Data type: {data_type:s} already set.')

    cls._data_types[data_type] = serializers

  @classmethod
  def RegisterDataTypes(cls, data_types):
    """Registers data types.

    Args:
      data_types (dict[str: dict[str, AttributeSerializer]]): attribute
          serializers with method per data types.

    Raises:
      KeyError: if the data type is already set.
    """
    for data_type, serializers in data_types.items():
      cls.RegisterDataType(data_type, serializers)



# -*- coding: utf-8 -*-
"""Attribute container JSON serializer."""

from acstore.containers import interface as containers_interface
from acstore.containers import manager as containers_manager
from acstore.helpers import schema as schema_helper


class AttributeContainerJSONSerializer(object):
  """Attribute container JSON serializer."""

  _CONTAINERS_MANAGER = containers_manager.AttributeContainersManager

  @classmethod
  def ConvertAttributeContainerToJSON(cls, attribute_container):
    """Converts an attribute container object into a JSON dictioary.

    The resulting dictionary of the JSON serialized objects consists of:
    {
        '__type__': 'AttributeContainer'
        '__container_type__': ...
        ...
    }

    Here '__type__' indicates the object base type. In this case
    'AttributeContainer'.

    '__container_type__' indicates the container type and rest of the elements
    of the dictionary that make up the attributes of the container.

    Args:
      attribute_container (AttributeContainer): attribute container.

    Returns:
      dict[str, object]: JSON serialized objects.
    """
    try:
      schema = cls._CONTAINERS_MANAGER.GetSchema(
          attribute_container.CONTAINER_TYPE)
    except ValueError:
      schema = {}

    json_dict = {
        '__type__': 'AttributeContainer',
        '__container_type__': attribute_container.CONTAINER_TYPE}

    for attribute_name, attribute_value in attribute_container.GetAttributes():
      data_type = schema.get(attribute_name, None)
      if data_type == 'AttributeContainerIdentifier' and isinstance(
          attribute_value, containers_interface.AttributeContainerIdentifier):
        attribute_value = attribute_value.CopyToString()
      else:
        serializer = schema_helper.SchemaHelper.GetAttributeSerializer(
            data_type, 'json')
        if serializer:
          attribute_value = serializer.SerializeValue(attribute_value)

      # JSON will not serialize certain runtime types like set, therefore
      # these are cast to list first.
      if isinstance(attribute_value, set):
        attribute_value = list(attribute_value)

      json_dict[attribute_name] = attribute_value

    return json_dict

  @classmethod
  def ConvertJSONToAttributeContainer(cls, json_dict):
    """Converts a JSON dictionary into an attribute container object.

    The dictionary of the JSON serialized objects consists of:
    {
        '__type__': 'AttributeContainer'
        '__container_type__': ...
        ...
    }

    Here '__type__' indicates the object base type. In this case
    'AttributeContainer'.

    '__container_type__' indicates the container type and rest of the elements
    of the dictionary that make up the attributes of the container.

    Args:
      json_dict (dict[str, object]): JSON serialized objects.

    Returns:
      AttributeContainer: attribute container.
    """
    # Use __container_type__ to indicate the attribute container type.
    container_type = json_dict.get('__container_type__', None)

    attribute_container = cls._CONTAINERS_MANAGER.CreateAttributeContainer(
        container_type)

    try:
      schema = cls._CONTAINERS_MANAGER.GetSchema(container_type)
    except ValueError:
      schema = {}

    supported_attribute_names = attribute_container.GetAttributeNames()
    for attribute_name, attribute_value in json_dict.items():
      if attribute_name in ('__container_type__', '__type__'):
        continue

      # Be strict about which attributes to set.
      if attribute_name not in supported_attribute_names:
        continue

      data_type = schema.get(attribute_name, None)
      if data_type == 'AttributeContainerIdentifier':
        identifier = containers_interface.AttributeContainerIdentifier()
        identifier.CopyFromString(attribute_value)
        attribute_value = identifier
      else:
        serializer = schema_helper.SchemaHelper.GetAttributeSerializer(
            data_type, 'json')
        if serializer:
          attribute_value = serializer.DeserializeValue(attribute_value)

      setattr(attribute_container, attribute_name, attribute_value)

    return attribute_container
